@@ -1,4 +1,4 @@
-# Copyright (C) 2015-2021  The Software Heritage developers
+# Copyright (C) 2015-2022  The Software Heritage developers
 # See the AUTHORS file at the top-level directory of this distribution
 # License: GNU General Public License version 3, or any later version
 # See top-level LICENSE file for more information
@@ -8,12 +8,12 @@ swh-storage.
 
 """
 from datetime import datetime
-from mmap import ACCESS_WRITE, mmap
+import difflib
 import os
 import pty
 import re
 import shutil
-from subprocess import Popen
+from subprocess import PIPE, Popen
 import tempfile
 from typing import Any, Dict, Iterator, List, Optional, Sequence, Tuple
 
@@ -42,6 +42,7 @@ from .utils import (
     OutputStream,
     init_svn_repo_from_archive_dump,
     init_svn_repo_from_dump,
+    svn_urljoin,
 )
 
 DEFAULT_BRANCH = b"HEAD"
@@ -141,7 +142,9 @@ Local repository not cleaned up for investigation: %s""",
             return
         self.svnrepo.clean_fs()
 
-    def swh_revision_hash_tree_at_svn_revision(self, revision: int) -> bytes:
+    def swh_revision_hash_tree_at_svn_revision(
+        self, revision: int
+    ) -> from_disk.Directory:
         """Compute and return the hash tree at a given svn revision.
 
         Args:
@@ -155,7 +158,7 @@ Local repository not cleaned up for investigation: %s""",
         local_dirname, local_url = self.svnrepo.export_temporary(revision)
         root_dir = from_disk.Directory.from_disk(path=local_url)
         self.svnrepo.clean_fs(local_dirname)
-        return root_dir.hash
+        return root_dir
 
     def _latest_snapshot_revision(
         self,
@@ -285,7 +288,9 @@ Local repository not cleaned up for investigation: %s""",
 
         return revision_start, revision_end
 
-    def _check_revision_divergence(self, rev: int, dir_id: bytes) -> None:
+    def _check_revision_divergence(
+        self, rev: int, dir_id: bytes, dir: from_disk.Directory
+    ) -> None:
         """Check for hash revision computation divergence.
 
            The Rationale behind this is that svn can trigger unknown edge cases (mixed
@@ -302,12 +307,75 @@ Local repository not cleaned up for investigation: %s""",
         """
 
         self.log.debug("Checking hash computations on revision %s...", rev)
-        checked_dir_id = self.swh_revision_hash_tree_at_svn_revision(rev)
+        checked_dir = self.swh_revision_hash_tree_at_svn_revision(rev)
+        checked_dir_id = checked_dir.hash
+
         if checked_dir_id != dir_id:
+            # do not bother checking tree differences if root directory id of reconstructed
+            # repository filesystem does not match the id of the one from the last loaded
+            # revision (can happen when called from post_load and tree differences were checked
+            # before the last revision to load)
+            if self.debug and dir_id == dir.hash:
+                for obj in checked_dir.iter_tree():
+                    path = obj.data["path"].replace(checked_dir.data["path"], b"")
+                    if not path:
+                        # ignore root directory
+                        continue
+                    if path not in dir:
+                        self.log.debug(
+                            "%s with path %s is missing in reconstructed repository filesystem",
+                            obj.object_type,  # type: ignore
+                            path,
+                        )
+                    elif dir[path].hash != checked_dir[path].hash:
+                        self.log.debug(
+                            "%s with path %s has different hash in reconstructed repository filesystem",  # noqa
+                            obj.object_type,  # type: ignore
+                            path,
+                        )
+                        if obj.object_type == "content":  # type: ignore
+                            self.log.debug(
+                                "expected sha1: %s, actual sha1: %s",
+                                hashutil.hash_to_hex(checked_dir[path].data["sha1"]),
+                                hashutil.hash_to_hex(dir[path].data["sha1"]),
+                            )
+                            # compute and display diff between contents
+                            file_path = (
+                                checked_dir[path]
+                                .data["path"]
+                                .replace(checked_dir.data["path"], b"")
+                            ).decode()
+                            with tempfile.TemporaryDirectory() as tmpdir:
+                                export_path = os.path.join(
+                                    tmpdir, os.path.basename(file_path)
+                                )
+                                assert self.svnrepo is not None
+                                self.svnrepo.export(
+                                    url=svn_urljoin(self.svnrepo.remote_url, file_path),
+                                    to=export_path,
+                                    rev=rev,
+                                    peg_rev=rev,
+                                    ignore_keywords=True,
+                                    overwrite=True,
+                                )
+                                with open(export_path, "rb") as exported_file, open(
+                                    dir[path].data["path"], "rb"
+                                ) as checkout_file:
+                                    diff_lines = difflib.diff_bytes(
+                                        difflib.unified_diff,
+                                        exported_file.read().split(b"\n"),
+                                        checkout_file.read().split(b"\n"),
+                                    )
+                                    self.log.debug(
+                                        "below is diff between files:\n"
+                                        + os.fsdecode(b"\n".join(list(diff_lines)[2:]))
+                                    )
+
             err = (
-                "Hash tree computation divergence detected "
+                "Hash tree computation divergence detected at revision %s "
                 "(%s != %s), stopping!"
                 % (
+                    rev,
                     hashutil.hash_to_hex(dir_id),
                     hashutil.hash_to_hex(checked_dir_id),
                 )
@@ -361,7 +429,7 @@ Local repository not cleaned up for investigation: %s""",
                 and self.check_revision != 0
                 and count % self.check_revision == 0
             ):
-                self._check_revision_divergence(rev, dir_id)
+                self._check_revision_divergence(rev, dir_id, root_directory)
 
             parents = (swh_revision.id,)
 
@@ -405,6 +473,7 @@ Local repository not cleaned up for investigation: %s""",
             local_dirname,
             self.max_content_size,
             self.from_dump,
+            debug=self.debug,
         )
 
         try:
@@ -536,6 +605,7 @@ Local repository not cleaned up for investigation: %s""",
             self._check_revision_divergence(
                 int(dict(self._last_revision.extra_headers)[b"svn_revision"]),
                 self._last_revision.directory,
+                self.svnrepo.swhreplay.directory,
             )
 
     def _create_tmp_dir(self, root_tmp_dir: str) -> str:
@@ -657,17 +727,19 @@ class SvnLoaderFromRemoteDump(SvnLoader):
             pass
         return svn_revision
 
-    def dump_svn_revisions(self, svn_url: str, last_loaded_svn_rev: int = -1) -> str:
-        """Generate a subversion dump file using the svnrdump tool. If the svnrdump
-        command failed somehow, the produced dump file is analyzed to determine if a
-        partial loading is still feasible.
+    def dump_svn_revisions(
+        self, svn_url: str, last_loaded_svn_rev: int = -1
+    ) -> Tuple[str, int]:
+        """Generate a compressed subversion dump file using the svnrdump tool and gzip.
+        If the svnrdump command failed somehow, the produced dump file is analyzed to
+        determine if a partial loading is still feasible.
 
         Raises:
             NotFound when the repository is no longer found at url
 
         Returns:
-            The dump_path of the repository mounted
-
+            The dump_path of the repository mounted and the max dumped revision number
+            (-1 if all revisions were dumped)
         """
         # Build the svnrdump command line
         svnrdump_cmd = ["svnrdump", "dump", svn_url]
@@ -684,12 +756,13 @@ class SvnLoaderFromRemoteDump(SvnLoader):
         # successfully dumped revision numbers are printed to it
         dump_temp_dir = tempfile.mkdtemp(dir=self.temp_dir)
         dump_name = "".join(c for c in svn_url if c.isalnum())
-        dump_path = "%s/%s.svndump" % (dump_temp_dir, dump_name)
+        dump_path = "%s/%s.svndump.gz" % (dump_temp_dir, dump_name)
         stderr_lines = []
         self.log.debug("Executing %s", " ".join(svnrdump_cmd))
         with open(dump_path, "wb") as dump_file:
+            gzip = Popen(["gzip"], stdin=PIPE, stdout=dump_file)
             stderr_r, stderr_w = pty.openpty()
-            svnrdump = Popen(svnrdump_cmd, stdout=dump_file, stderr=stderr_w)
+            svnrdump = Popen(svnrdump_cmd, stdout=gzip.stdin, stderr=stderr_w)
             os.close(stderr_w)
             stderr_stream = OutputStream(stderr_r)
             readable = True
@@ -706,9 +779,12 @@ class SvnLoaderFromRemoteDump(SvnLoader):
                         error_messages.append(line)
             svnrdump.wait()
             os.close(stderr_r)
+            # denote end of read file
+            gzip.stdin.close()
+            gzip.wait()
 
         if svnrdump.returncode == 0:
-            return dump_path
+            return dump_path, -1
 
         # There was an error but it does not mean that no revisions
         # can be loaded.
@@ -736,27 +812,8 @@ class SvnLoaderFromRemoteDump(SvnLoader):
                     last_loaded_svn_rev + 1,
                     last_dumped_rev,
                 )
-                # Truncate the dump file after the last successfully dumped
-                # revision to avoid the loading of corrupted data
-                self.log.debug(
-                    (
-                        "Truncating dump file after the last "
-                        "successfully dumped revision (%s) to avoid "
-                        "the loading of corrupted data"
-                    ),
-                    last_dumped_rev,
-                )
-
-                with open(dump_path, "r+b") as f:
-                    with mmap(f.fileno(), 0, access=ACCESS_WRITE) as s:
-                        pattern = (
-                            "Revision-number: %s" % (last_dumped_rev + 1)
-                        ).encode()
-                        n = s.rfind(pattern)
-                        if n != -1:
-                            s.resize(n)
                 self.truncated_dump = True
-                return dump_path
+                return dump_path, last_dumped_rev
             elif last_dumped_rev != -1 and last_dumped_rev < last_loaded_svn_rev:
                 raise Exception(
                     (
@@ -783,7 +840,11 @@ class SvnLoaderFromRemoteDump(SvnLoader):
         last_loaded_svn_rev = self.get_last_loaded_svn_rev(self.origin.url)
 
         self.svnrepo = self.svn_repo(
-            self.origin.url, self.origin.url, self.temp_dir, self.max_content_size
+            self.origin.url,
+            self.origin.url,
+            self.temp_dir,
+            self.max_content_size,
+            debug=self.debug,
         )
 
         # Ensure to use remote URL retrieved by SvnRepo as origin URL might redirect
@@ -809,7 +870,7 @@ class SvnLoaderFromRemoteDump(SvnLoader):
 
         # Then try to generate a dump file containing relevant svn revisions
         # to load, an exception will be thrown if something wrong happened
-        dump_path = self.dump_svn_revisions(self.svn_url, last_loaded_svn_rev)
+        dump_path, max_rev = self.dump_svn_revisions(self.svn_url, last_loaded_svn_rev)
 
         # Finally, mount the dump and load the repository
         self.log.debug('Mounting dump file with "svnadmin load".')
@@ -818,6 +879,8 @@ class SvnLoaderFromRemoteDump(SvnLoader):
             prefix=TEMPORARY_DIR_PREFIX_PATTERN,
             suffix="-%s" % os.getpid(),
             root_dir=self.temp_dir,
+            gzip=True,
+            max_rev=max_rev,
         )
         self.svn_url = "file://%s" % self.repo_path
         super().prepare()
