@@ -11,6 +11,7 @@ from __future__ import annotations
 
 import codecs
 from collections import defaultdict
+from copy import copy
 from dataclasses import dataclass, field
 from distutils.dir_util import copy_tree
 from itertools import chain
@@ -33,7 +34,7 @@ from typing import (
 )
 
 import click
-from subvertpy import SubversionException, delta, properties
+from subvertpy import SubversionException, properties
 from subvertpy.ra import Auth, RemoteAccess, get_username_provider
 
 from swh.model import from_disk, hashutil
@@ -49,99 +50,9 @@ from swh.loader.svn.utils import (
     svn_urljoin,
 )
 
-_eol_style = {"native": b"\n", "CRLF": b"\r\n", "LF": b"\n", "CR": b"\r"}
-
 logger = logging.getLogger(__name__)
 
 
-def _normalize_line_endings(lines: bytes, eol_style: str = "native") -> bytes:
-    r"""Normalize line endings to unix (\\n), windows (\\r\\n) or mac (\\r).
-
-    Args:
-        lines: The lines to normalize
-
-        eol_style: The line ending format as defined for
-            svn:eol-style property. Acceptable values are 'native',
-            'CRLF', 'LF' and 'CR'
-
-    Returns:
-        Lines with endings normalized
-    """
-    if eol_style in _eol_style:
-        lines = lines.replace(_eol_style["CRLF"], _eol_style["LF"]).replace(
-            _eol_style["CR"], _eol_style["LF"]
-        )
-        if _eol_style[eol_style] != _eol_style["LF"]:
-            lines = lines.replace(_eol_style["LF"], _eol_style[eol_style])
-
-    return lines
-
-
-def apply_txdelta_handler(
-    sbuf: bytes, target_stream: BinaryIO
-) -> Callable[[Any, bytes, BinaryIO], None]:
-    """Return a function that can be called repeatedly with txdelta windows.
-    When done, closes the target_stream.
-
-    Adapted from subvertpy.delta.apply_txdelta_handler to close the
-    stream when done.
-
-    Args:
-        sbuf: Source buffer
-        target_stream: Target stream to write to.
-
-    Returns:
-        Function to be called to apply txdelta windows
-
-    """
-
-    def apply_window(
-        window: Any, sbuf: bytes = sbuf, target_stream: BinaryIO = target_stream
-    ):
-        if window is None:
-            target_stream.close()
-            return  # Last call
-        patch = delta.apply_txdelta_window(sbuf, window)
-        target_stream.write(patch)
-
-    return apply_window
-
-
-def read_svn_link(data: bytes) -> Tuple[bytes, bytes]:
-    """Read the svn link's content.
-
-    Args:
-        data: svn link's raw content
-
-    Returns:
-        The tuple of (filetype, destination path)
-
-    """
-    split_byte = b" "
-    first_line = data.split(b"\n")[0]
-    filetype, *src = first_line.split(split_byte)
-    target = split_byte.join(src)
-    return filetype, target
-
-
-def is_file_an_svnlink_p(fullpath: bytes) -> Tuple[bool, bytes]:
-    """Determine if a filepath is an svnlink or something else.
-
-    Args:
-        fullpath: Full path to the potential symlink to check
-
-    Returns:
-        Tuple containing a boolean value to determine if it's indeed a symlink
-        (as per svn) and the link target.
-
-    """
-    if os.path.islink(fullpath):
-        return False, b""
-    with open(fullpath, "rb") as f:
-        filetype, src = read_svn_link(f.read())
-        return filetype == b"link", src
-
-
 def _ra_codecs_error_handler(e: UnicodeError) -> Tuple[Union[str, bytes], int]:
     """Subvertpy may fail to decode to utf-8 the user svn properties.  As
        they are not used by the loader, return an empty string instead
@@ -161,25 +72,6 @@ NOEXEC_FLAG = 2
 SVN_PROPERTY_EOL = "svn:eol-style"
 
 
-@dataclass
-class FileState:
-    """Persists some file states (eg. end of lines style) across revisions while
-    replaying them."""
-
-    eol_style: Optional[str] = None
-    """EOL state check mess"""
-
-    svn_special_path_non_link_data: Optional[bytes] = None
-    """keep track of non link file content with svn:special property set"""
-
-    # default value: 0, 1: set the flag, 2: remove the exec flag
-    executable: int = DEFAULT_FLAG
-    """keep track if file is executable when setting svn:executable property"""
-
-    link: bool = False
-    """keep track if file is a svn link when setting svn:special property"""
-
-
 class FileEditor:
     """File Editor in charge of updating file on disk and memory objects."""
 
@@ -199,155 +91,47 @@ class FileEditor:
         directory: from_disk.Directory,
         rootpath: bytes,
         path: bytes,
-        state: FileState,
         svnrepo: SvnRepo,
     ):
         self.directory = directory
         self.path = path
         self.fullpath = os.path.join(rootpath, path)
-        self.state = state
         self.svnrepo = svnrepo
-        self.editor = svnrepo.swhreplay.editor
+        self.editor: Editor = svnrepo.swhreplay.editor
 
     def change_prop(self, key: str, value: str) -> None:
-        if key == properties.PROP_EXECUTABLE:
-            if value is None:  # bit flip off
-                self.state.executable = NOEXEC_FLAG
-            else:
-                self.state.executable = EXEC_FLAG
-        elif key == properties.PROP_SPECIAL:
-            # Possibly a symbolic link. We cannot check further at
-            # that moment though, patch(s) not being applied yet
-            self.state.link = value is not None
-        elif key == SVN_PROPERTY_EOL:
-            # backup end of line style for file
-            self.state.eol_style = value
-
-    def __make_symlink(self, src: bytes) -> None:
-        """Convert the svnlink to a symlink on disk.
-
-        This function expects self.fullpath to be a svn link.
-
-        Args:
-            src: Path to the link's source
-
-        Return:
-            tuple: The svnlink's data tuple:
-
-                - type (should be only 'link')
-                - <path-to-src>
-
-        """
-        os.remove(self.fullpath)
-        os.symlink(src=src, dst=self.fullpath)
-
-    def __make_svnlink(self) -> bytes:
-        """Convert the symlink to a svnlink on disk.
-
-        Return:
-            The symlink's svnlink data (``b'type <path-to-src>'``)
-
-        """
-        # we replace the symlink by a svnlink
-        # to be able to patch the file on future commits
-        src = os.readlink(self.fullpath)
-        os.remove(self.fullpath)
-        sbuf = b"link " + src
-        with open(self.fullpath, "wb") as f:
-            f.write(sbuf)
-        return sbuf
+        if self.editor.debug:
+            logger.debug(
+                "Setting property %s to value %s on path %s", key, value, self.path
+            )
 
     def apply_textdelta(self, base_checksum) -> Callable[[Any, bytes, BinaryIO], None]:
-        # if the filepath matches an external, do not apply local patch
-        if self.path in self.editor.external_paths:
-            return lambda *args: None
-
-        if os.path.lexists(self.fullpath):
-            if os.path.islink(self.fullpath):
-                # svn does not deal with symlink so we transform into
-                # real svn symlink for potential patching in later
-                # commits
-                sbuf = self.__make_svnlink()
-                self.state.link = True
-            else:
-                with open(self.fullpath, "rb") as f:
-                    sbuf = f.read()
-        else:
-            sbuf = b""
-
-        t = open(self.fullpath, "wb")
-        return apply_txdelta_handler(sbuf, target_stream=t)
+        if self.editor.debug:
+            logger.debug("Applying textdelta to file %s", self.path)
+        # do not apply textdelta, file will be fully exported when closing the editor
+        return lambda *args: None
 
     def close(self) -> None:
-        """When done with the file, this is called.
-
-        So the file exists and is updated, we can:
-
-        - adapt accordingly its execution flag if any
-        - compute the objects' checksums
-        - replace the svnlink with a real symlink (for disk
-          computation purposes)
+        """When done with a file added or modified in the current replayed revision,
+        we export it to disk and update the from_disk model.
 
         """
+        if self.editor.debug:
+            logger.debug("Closing file %s", self.path)
 
-        if self.state.link:
-            # can only check now that the link is a real one
-            # since patch has been applied
-            is_link, src = is_file_an_svnlink_p(self.fullpath)
-            if is_link:
-                self.__make_symlink(src)
-            elif not os.path.isdir(self.fullpath):  # not a real link ...
-                # when a file with the svn:special property set is not a svn link,
-                # the svn export operation might extract a truncated version of it
-                # if it is a binary file, so ensure to produce the same file as the
-                # export operation.
-                with open(self.fullpath, "rb") as f:
-                    content = f.read()
-                self.svnrepo.export(
-                    os.path.join(self.svnrepo.remote_url.encode(), self.path),
-                    to=self.fullpath,
-                    peg_rev=self.editor.revnum,
-                    ignore_keywords=True,
-                    overwrite=True,
-                )
-                with open(self.fullpath, "rb") as f:
-                    exported_data = f.read()
-                    if exported_data != content:
-                        # keep track of original file content in order to restore
-                        # it if the svn:special property gets unset in another revision
-                        self.state.svn_special_path_non_link_data = content
-        elif os.path.islink(self.fullpath):
-            # path was a symbolic link in previous revision but got the property
-            # svn:special unset in current one, revert its content to svn link format
-            self.__make_svnlink()
-        elif self.state.svn_special_path_non_link_data is not None:
-            # path was a non link file with the svn:special property previously set
-            # and got truncated on export, restore its original content
-            with open(self.fullpath, "wb") as f:
-                f.write(self.state.svn_special_path_non_link_data)
-                self.state.svn_special_path_non_link_data = None
-
-        is_link = os.path.islink(self.fullpath)
-        if not is_link:  # if a link, do nothing regarding flag
-            if self.state.executable == EXEC_FLAG:
-                os.chmod(self.fullpath, 0o755)
-            elif self.state.executable == NOEXEC_FLAG:
-                os.chmod(self.fullpath, 0o644)
+        if self.path not in self.editor.external_paths:
+            # export file to disk if its path does not match an external
+            self.svnrepo.export(
+                os.path.join(self.svnrepo.remote_url, os.fsdecode(self.path)),
+                to=self.fullpath,
+                rev=self.editor.revnum,
+                peg_rev=self.editor.revnum,
+                ignore_keywords=True,
+                overwrite=True,
+            )
 
         # And now compute file's checksums
-        if self.state.eol_style and not is_link:
-            # ensure to normalize line endings as defined by svn:eol-style
-            # property to get the same file checksum as after an export
-            # or checkout operation with subversion
-            with open(self.fullpath, "rb") as f:
-                data = f.read()
-                data = _normalize_line_endings(data, self.state.eol_style)
-                mode = os.lstat(self.fullpath).st_mode
-                self.directory[self.path] = from_disk.Content.from_bytes(
-                    mode=mode, data=data
-                )
-        else:
-            self.directory[self.path] = from_disk.Content.from_file(path=self.fullpath)
+        self.directory[self.path] = from_disk.Content.from_file(path=self.fullpath)
 
 
 ExternalDefinition = Tuple[str, Optional[int], bool]
@@ -361,6 +145,8 @@ class DirState:
     externals: Dict[str, List[ExternalDefinition]] = field(default_factory=dict)
     """Map a path in the directory to a list of (external_url, revision, relative_url)
     targeting it"""
+    externals_paths: Set[bytes] = field(default_factory=set)
+    """Keep track of all external paths reachable from the directory"""
 
 
 class DirEditor:
@@ -374,7 +160,6 @@ class DirEditor:
         "directory",
         "rootpath",
         "path",
-        "file_states",
         "dir_states",
         "svnrepo",
         "editor",
@@ -386,7 +171,6 @@ class DirEditor:
         directory: from_disk.Directory,
         rootpath: bytes,
         path: bytes,
-        file_states: Dict[bytes, FileState],
         dir_states: Dict[bytes, DirState],
         svnrepo: SvnRepo,
     ):
@@ -395,7 +179,6 @@ class DirEditor:
         self.path = path
         # build directory on init
         os.makedirs(rootpath, exist_ok=True)
-        self.file_states = file_states
         self.dir_states = dir_states
         self.svnrepo = svnrepo
         self.editor = svnrepo.swhreplay.editor
@@ -413,11 +196,8 @@ class DirEditor:
             path: to remove from the current objects.
 
         """
-        try:
+        if path in self.directory:
             entry_removed = self.directory[path]
-        except KeyError:
-            entry_removed = None
-        else:
             del self.directory[path]
             fpath = os.path.join(self.rootpath, path)
             if isinstance(entry_removed, from_disk.Directory):
@@ -425,67 +205,124 @@ class DirEditor:
             else:
                 os.remove(fpath)
 
-        # when deleting a directory ensure to remove any svn property for the
-        # file it contains as they can be added again later in another revision
-        # without the same property set
-        fullpath = os.path.join(self.rootpath, path)
-        for state_path in list(self.file_states):
-            if state_path.startswith(fullpath + b"/"):
-                del self.file_states[state_path]
-
     def open_directory(self, path: str, *args) -> DirEditor:
         """Updating existing directory."""
+        if self.editor.debug:
+            logger.debug("Opening directory %s", path)
         return DirEditor(
             self.directory,
             rootpath=self.rootpath,
             path=os.fsencode(path),
-            file_states=self.file_states,
             dir_states=self.dir_states,
             svnrepo=self.svnrepo,
         )
 
-    def add_directory(self, path: str, *args) -> DirEditor:
+    def add_directory(
+        self, path: str, copyfrom_path: Optional[str] = None, copyfrom_rev: int = -1
+    ) -> DirEditor:
         """Adding a new directory."""
-        path_bytes = os.fsencode(path)
+        if self.editor.debug:
+            logger.debug(
+                "Adding directory %s, copyfrom_path = %s, copyfrom_rev = %s",
+                path,
+                copyfrom_path,
+                copyfrom_rev,
+            )
 
-        os.makedirs(os.path.join(self.rootpath, path_bytes), exist_ok=True)
-        if path_bytes and path_bytes not in self.directory:
-            self.dir_states[path_bytes] = DirState()
-            self.directory[path_bytes] = from_disk.Directory()
+        path_bytes = os.fsencode(path)
+        fullpath = os.path.join(self.rootpath, path_bytes)
+
+        os.makedirs(fullpath, exist_ok=True)
+        if copyfrom_rev == -1:
+            if path_bytes and path_bytes not in self.directory:
+                self.directory[path_bytes] = from_disk.Directory()
+        else:
+            url = svn_urljoin(self.svnrepo.remote_url, copyfrom_path)
+            self.remove_child(path_bytes)
+            self.svnrepo.export(
+                url,
+                to=fullpath,
+                peg_rev=copyfrom_rev,
+                ignore_keywords=True,
+                overwrite=True,
+            )
+            self.directory[path_bytes] = from_disk.Directory.from_disk(path=fullpath)
+
+            assert copyfrom_path is not None
+            copyfrom_path_bytes = os.fsencode(copyfrom_path).lstrip(b"/")
+            copyfrom_fullpath = os.path.join(self.rootpath, copyfrom_path_bytes)
+
+            def _copy_dir_state(path: bytes, copied_path: bytes):
+                self.dir_states[path] = copy(self.dir_states[copied_path])
+                for external_path in self.dir_states[path].externals_paths:
+                    self.editor.external_paths[os.path.join(path, external_path)] += 1
+
+            _copy_dir_state(path_bytes, copyfrom_path_bytes)
+            for root, dirs, _ in os.walk(fullpath):
+                for dir in dirs:
+                    dir_fullpath = os.path.join(root, dir)
+                    copied_dir_fullpath = dir_fullpath.replace(
+                        fullpath, copyfrom_fullpath
+                    )
+                    dir_path = dir_fullpath.replace(self.rootpath, b"").lstrip(b"/")
+                    copied_dir_path = copied_dir_fullpath.replace(self.rootpath, b"")
+                    _copy_dir_state(dir_path, copied_dir_path.lstrip(b"/"))
 
         return DirEditor(
             self.directory,
             self.rootpath,
             path_bytes,
-            self.file_states,
             self.dir_states,
             svnrepo=self.svnrepo,
         )
 
     def open_file(self, path: str, *args) -> FileEditor:
         """Updating existing file."""
+        if self.editor.debug:
+            logger.debug("Opening file %s", path)
+
         path_bytes = os.fsencode(path)
         self.directory[path_bytes] = from_disk.Content()
-        fullpath = os.path.join(self.rootpath, path_bytes)
         return FileEditor(
             self.directory,
             rootpath=self.rootpath,
             path=path_bytes,
-            state=self.file_states[fullpath],
             svnrepo=self.svnrepo,
         )
 
-    def add_file(self, path: str, *args) -> FileEditor:
+    def add_file(
+        self, path: str, copyfrom_path: Optional[str] = None, copyfrom_rev: int = -1
+    ) -> FileEditor:
         """Creating a new file."""
+        if self.editor.debug:
+            logger.debug(
+                "Adding file %s, copyfrom_path = %s, copyfrom_rev = %s",
+                path,
+                copyfrom_path,
+                copyfrom_rev,
+            )
+
         path_bytes = os.fsencode(path)
-        self.directory[path_bytes] = from_disk.Content()
         fullpath = os.path.join(self.rootpath, path_bytes)
-        self.file_states[fullpath] = FileState()
+
+        if copyfrom_rev == -1:
+            self.directory[path_bytes] = from_disk.Content()
+        else:
+            url = svn_urljoin(self.svnrepo.remote_url, copyfrom_path)
+            self.remove_child(path_bytes)
+            self.svnrepo.export(
+                url,
+                to=fullpath,
+                peg_rev=copyfrom_rev,
+                ignore_keywords=True,
+                overwrite=True,
+            )
+            self.directory[path_bytes] = from_disk.Content.from_file(path=fullpath)
+
         return FileEditor(
             self.directory,
             self.rootpath,
             path_bytes,
-            state=self.file_states[fullpath],
             svnrepo=self.svnrepo,
         )
 
@@ -541,17 +378,43 @@ class DirEditor:
 
     def delete_entry(self, path: str, revision: int) -> None:
         """Remove a path."""
+        if self.editor.debug:
+            logger.debug("Deleting directory entry %s", path)
+
         path_bytes = os.fsencode(path)
+        fullpath = os.path.join(self.rootpath, path_bytes)
+
+        if os.path.isdir(fullpath):
+            # remove all external paths associated to the removed directory
+            # (we cannot simply remove a root external directory as externals
+            # paths associated to ancestor directories can overlap)
+            for external_path in self.dir_states[path_bytes].externals_paths:
+                self.remove_external_path(
+                    external_path,
+                    root_path=path_bytes,
+                    remove_subpaths=False,
+                    force=True,
+                )
+
         if path_bytes not in self.editor.external_paths:
-            fullpath = os.path.join(self.rootpath, path_bytes)
-            self.file_states.pop(fullpath, None)
             self.remove_child(path_bytes)
+        elif os.path.isdir(fullpath):
+            # versioned and external paths can overlap so we need to iterate on
+            # all subpaths to check which ones to remove
+            for root, dirs, files in os.walk(fullpath):
+                for p in chain(dirs, files):
+                    full_repo_path = os.path.join(root, p)
+                    repo_path = full_repo_path.replace(self.rootpath + b"/", b"")
+                    if repo_path not in self.editor.external_paths:
+                        self.remove_child(repo_path)
 
     def close(self):
         """Function called when we finish processing a repository.
 
         SVN external definitions are processed by it.
         """
+        if self.editor.debug:
+            logger.debug("Closing directory %s", self.path)
 
         prev_externals = self.dir_states[self.path].externals
 
@@ -714,9 +577,6 @@ class DirEditor:
             # copy exported path to reconstructed filesystem
             fullpath = os.path.join(self.rootpath, dest_fullpath)
 
-            # update from_disk model and store external paths
-            self.editor.external_paths[dest_fullpath] += 1
-
             if os.path.isfile(temp_path):
                 if os.path.islink(fullpath):
                     # remove destination file if it is a link
@@ -752,37 +612,56 @@ class DirEditor:
                 self.directory[dest_fullpath] = from_disk.Directory.from_disk(
                     path=fullpath
                 )
-                external_paths = set()
-                for root, dirs, files in os.walk(fullpath):
-                    external_paths.update(
-                        [
-                            os.path.join(root.replace(self.rootpath + b"/", b""), p)
-                            for p in chain(dirs, files)
-                        ]
-                    )
-                for external_path in external_paths:
-                    self.editor.external_paths[external_path] += 1
+
+            # update set of external paths reachable from the directory
+            external_paths = set()
+            dest_path_part = dest_path.split(b"/")
+            for i in range(1, len(dest_path_part) + 1):
+                external_paths.add(b"/".join(dest_path_part[:i]))
+
+            for root, dirs, files in os.walk(temp_path):
+                external_paths.update(
+                    [
+                        os.path.join(
+                            dest_path,
+                            os.path.join(root, p).replace(temp_path, b"").strip(b"/"),
+                        )
+                        for p in chain(dirs, files)
+                    ]
+                )
+
+            self.dir_states[self.path].externals_paths.update(external_paths)
+
+            for external_path in external_paths:
+                self.editor.external_paths[os.path.join(self.path, external_path)] += 1
 
             # ensure hash update for the directory with externals set
             self.directory[self.path].update_hash(force=True)
 
     def remove_external_path(
-        self, external_path: bytes, remove_subpaths: bool = True, force: bool = False
+        self,
+        external_path: bytes,
+        remove_subpaths: bool = True,
+        force: bool = False,
+        root_path: Optional[bytes] = None,
     ) -> None:
         """Remove a previously exported SVN external path from
         the reconstructed filesystem.
         """
-        fullpath = os.path.join(self.path, external_path)
+        path = root_path if root_path else self.path
+        fullpath = os.path.join(path, external_path)
+
+        if self.editor.debug:
+            logger.debug("Removing external path %s", fullpath)
 
         # decrement number of references for external path when we really remove it
         # (when remove_subpaths is False, we just cleanup the external path before
         # copying exported paths in it)
-        if fullpath in self.editor.external_paths and remove_subpaths:
+        if force or (fullpath in self.editor.external_paths and remove_subpaths):
             self.editor.external_paths[fullpath] -= 1
 
         if (
-            force
-            or fullpath in self.editor.external_paths
+            fullpath in self.editor.external_paths
             and self.editor.external_paths[fullpath] == 0
         ):
             self.remove_child(fullpath)
@@ -795,10 +674,10 @@ class DirEditor:
                         self.editor.external_paths.pop(path)
 
         if remove_subpaths:
-            subpath_split = external_path.split(b"/")[:-1]
+            subpath_split = fullpath.split(b"/")[:-1]
             for i in reversed(range(1, len(subpath_split) + 1)):
                 # delete external sub-directory only if it is not versioned
-                subpath = os.path.join(self.path, b"/".join(subpath_split[0:i]))
+                subpath = b"/".join(subpath_split[0:i])
                 try:
                     self.svnrepo.client.info(
                         svn_urljoin(self.svnrepo.remote_url, os.fsdecode(subpath)),
@@ -843,10 +722,10 @@ class Editor:
         directory: from_disk.Directory,
         svnrepo: SvnRepo,
         temp_dir: str,
+        debug: bool = False,
     ):
         self.rootpath = rootpath
         self.directory = directory
-        self.file_states: Dict[bytes, FileState] = defaultdict(FileState)
         self.dir_states: Dict[bytes, DirState] = defaultdict(DirState)
         self.external_paths: Dict[bytes, int] = defaultdict(int)
         self.valid_externals: Dict[bytes, Tuple[str, bool]] = {}
@@ -855,6 +734,7 @@ class Editor:
         self.externals_cache: Dict[ExternalDefinition, bytes] = {}
         self.svnrepo = svnrepo
         self.revnum = None
+        self.debug = debug
 
     def set_target_revision(self, revnum) -> None:
         self.revnum = revnum
@@ -870,7 +750,6 @@ class Editor:
             self.directory,
             rootpath=self.rootpath,
             path=b"",
-            file_states=self.file_states,
             dir_states=self.dir_states,
             svnrepo=self.svnrepo,
         )
@@ -886,6 +765,7 @@ class Replay:
         svnrepo: SvnRepo,
         temp_dir: str,
         directory: Optional[from_disk.Directory] = None,
+        debug: bool = False,
     ):
         self.conn = conn
         self.rootpath = rootpath
@@ -893,10 +773,14 @@ class Replay:
             directory = from_disk.Directory()
         self.directory = directory
         self.editor = Editor(
-            rootpath=rootpath, directory=directory, svnrepo=svnrepo, temp_dir=temp_dir
+            rootpath=rootpath,
+            directory=directory,
+            svnrepo=svnrepo,
+            temp_dir=temp_dir,
+            debug=debug,
         )
 
-    def replay(self, rev: int) -> from_disk.Directory:
+    def replay(self, rev: int, low_water_mark: int) -> from_disk.Directory:
         """Replay svn actions between rev and rev+1.
 
         This method updates in place the self.editor.directory, as well as the
@@ -907,12 +791,12 @@ class Replay:
 
         """
         codecs.register_error("strict", _ra_codecs_error_handler)
-        self.conn.replay(rev, rev + 1, self.editor)
+        self.conn.replay(rev, low_water_mark, self.editor)
         codecs.register_error("strict", codecs.strict_errors)
         return self.editor.directory
 
     def compute_objects(
-        self, rev: int
+        self, rev: int, low_water_mark: int
     ) -> Tuple[List[Content], List[SkippedContent], List[Directory]]:
         """Compute objects added or modified at revisions rev.
         Expects the state to be at previous revision's objects.
@@ -925,7 +809,7 @@ class Replay:
             mutates the filesystem at rootpath accordingly.
 
         """
-        self.replay(rev)
+        self.replay(rev, low_water_mark)
 
         contents: List[Content] = []
         skipped_contents: List[SkippedContent] = []
